#include "config.h"

#include "ds485-socket-server.h"

#include <assert.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>

#include <utlist.h>

#include <digitalSTROM/ds485.h>
#include <digitalSTROM/ds485-core.h>
#include "ds485-socket-interface.h"

#include "ds485-socket-server-impl.h"

#include "base/log.h"
#include "base/util.h"

#ifndef POLLRDHUP
#define POLLRDHUP 0
#endif

#define SVCBUFLEN 8500

static void ds485_server_process_client_data(dS485dClient *cl);
static void ds485_server_free_client(dS485dClient *cl);

static int update_pretty_name(dS485dClient *cl)
{
    struct sockaddr_in fsin;
    int rval;

    rval = get_peername(cl->fd, &fsin);
    if (rval < 0) {
        // already logged
        return rval;
    }

    /* increase peer_name size if doing changes */
    rval = snprintf(cl->peer_name, sizeof(cl->peer_name), "%.15s:%05d.%02d",
                    inet_ntoa(fsin.sin_addr),
                    ntohs(fsin.sin_port),
                    cl->transactionId);

    if (rval >= (int)sizeof(cl->peer_name)) {
        report(LOG_NOTICE, cl->peer_name, "snprintf truncated\n");
        /* minor issue, do not refuse client */
        return 0;
    }

    return 0;
}

/* CAUTION: requires cl->lock to be held */
void _ds485n_client_detach(dS485dClient *cl)
{
    if (-1 == cl->fd) {
        return;
    }

    report2(LOG_DEBUG, cl->peer_name, "detaching\n");
    close(cl->fd);
    cl->fd = -1;
    cl->ts_detach = time(NULL);
}

static void clDetach(dS485dClient* cl)
{
    pthread_mutex_lock(&cl->lock);   //xx
    _ds485n_client_detach(cl);
    pthread_mutex_unlock(&cl->lock);
}

/* CAUTION: clientListRWLock must be held */
static int _ds485_server_clients_count(struct dS485_server *server)
{
    dS485dClient* ign;
    int count = 0;

    LL_FOREACH(server->clSessions, ign) {
        count++;
    }
    return count;
}

static int ds485_server_append_client(struct dS485_server *server,
                                      dS485dClient* cl)
{
    pthread_rwlock_wrlock(&server->clientListRWLock);

    if (server->client_count >= MAX_CLIENTS) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report2(LOG_INFO, cl->peer_name,
                "transaction id pool depleted, rejecting client");
        return -EBUSY;
    }

    if (_ds485_server_clients_count(server) != server->client_count) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report(LOG_ERR, "append: client count mismatch\n");
        return -ERANGE;
    }

    cl->server = server;
    cl->transactionId = server->transid_pool[server->client_count++];
    LL_PREPEND(server->clSessions, cl);
    pthread_rwlock_unlock(&server->clientListRWLock);

    (void)update_pretty_name(cl);
    return 0;
}

static int ds485_server_remove_client(struct dS485_server *server,
                                      dS485dClient* cl)
{
    pthread_rwlock_wrlock(&server->clientListRWLock);

    if (_ds485_server_clients_count(server) != server->client_count) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report(LOG_ERR, "remove: client count mismatch\n");
        return -ERANGE;
    }

    if (server->client_count <= 0) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report(LOG_ERR, "remove: client not registered\n");
        return -EINVAL;
    }

    /*
     * TODO: remove client is only called from the dispatch thread.
     * If we ever remove clients when message send fails, we need
     * a better check if the client has already been removed
     * from the list
     */

    server->transid_pool[--server->client_count] = cl->transactionId;
    LL_DELETE(server->clSessions, cl);
    pthread_rwlock_unlock(&server->clientListRWLock);

    return 0;
}

static void cleanup_helper(dS485dClient *cl)
{
    ds485_server_remove_client(cl->server, cl);
    ds485_server_free_client(cl);
}

static void* sysServiceThread(void* arg)
{
    dS485dClient *cl = arg;
    struct dS485_server *server = cl->server;

    struct timeval tv;
    fd_set rfds;

    int rval;

    report2(LOG_NOTICE, cl->peer_name, "new client connection %d\n",
            cl->transactionId);

    pthread_cleanup_push(cleanup_helper, cl);

    /* handle client connection */
    while (server->active && cl->fd > 0) {

        cl->ts_activity = time(NULL);

        FD_ZERO(&rfds);
        FD_SET(cl->fd, &rfds);

        tv.tv_sec = MAX_THREAD_SLEEP;
        tv.tv_usec = 0;
        rval = select(cl->fd + 1, &rfds, NULL, NULL, &tv);
        if (rval < 0) {
            if (errno == EINTR)
                continue;
            report2(LOG_WARNING, cl->peer_name, "error %d [%d]\n", rval);
            clDetach(cl);
            break;
        }
        else if (rval == 0) {
            /* timeout */
            continue;
        }
        if (FD_ISSET(cl->fd, &rfds)) {
            rval = recv(cl->fd, cl->rxBuffer + cl->rxLen, SVCBUFLEN - cl->rxLen, 0);
            if (rval < 0) {
                report2(LOG_WARNING, cl->peer_name, "read: %s(%d)\n",
                        strerror(errno), errno);
                if (errno == ECONNRESET) {
                    clDetach(cl);
                }
                continue;
            }
            else if (rval == 0) {
                report2(LOG_WARNING, cl->peer_name, "peer closed connection\n");
                clDetach(cl);
                break;
            }
            else {
                cl->rxLen += rval;
            }
        }
        else {
            continue;
        }
        ds485_server_process_client_data(cl);

    }

    report2(LOG_NOTICE, cl->peer_name, "connection terminated\n");
    pthread_cleanup_pop(1);
    return NULL;
}

/**
 * ds485_server_init_client
 * @sock -- client socket
 */
static dS485dClient* ds485_server_create_client(int cfd)
{
    dS485dClient* cls;
    int opts, rval;

    opts = fcntl(cfd, F_GETFL);
    if (opts < 0) {
        report(LOG_WARNING, "new client: F_GETFL failed %s", strerror(errno));
        close(cfd);
        return NULL;
    }

    if (fcntl(cfd, F_SETFL, opts | O_NONBLOCK) < 0) {
        report(LOG_WARNING, "new client: setting O_NONBLOCK failed %s", strerror(errno));
        close(cfd);
        return NULL;
    }

    cls = calloc(sizeof(dS485dClient), 1);
    if (cls == NULL) {
        report(LOG_WARNING, "new client: %s", strerror(ENOMEM));
        close(cfd);
        return NULL;
    }

    cls->fd = cfd;
    cls->transactionId = -1; /* 0 is valid id */
    pthread_mutex_init(&cls->lock, NULL);

    rval = update_pretty_name(cls);
    if (rval) {
        report(LOG_ERR, "new client: failed to get peername %s\n", strerror(-rval));
        close(cfd);
        pthread_mutex_destroy(&cls->lock);
        free(cls);
        return NULL;
    }

    cls->rxBuffer = malloc(SVCBUFLEN);
    if (cls->rxBuffer == NULL) {
        report(LOG_ERR, "new client: failed to allocate recevice buffer\n");
        close(cfd);
        pthread_mutex_destroy(&cls->lock);
        free(cls);
        return NULL;
    }

    return cls;
}

static void ds485_server_free_client(dS485dClient *cl)
{
    clDetach(cl);
    pthread_mutex_destroy(&cl->lock);
    free(cl->rxBuffer);
    free(cl);
}

static void ds485_server_process_client_data(dS485dClient *cl)
{
    uint8_t  *p;

    p = cl->rxBuffer;
    while (cl->rxLen >= 3) {
        int pkt_len = p[2] << 8 | p[1];

        /* not enough data */
        if (pkt_len > cl->rxLen) {
            report2(LOG_DEBUG, cl->peer_name, "not enough data, retry (%d > %d)\n",
                    pkt_len, cl->rxLen);
            break;
        }

        /* Process packet data ... */
        ds485_netlib_packet_parser(cl, p);
        /* if (rval) TODO */

        p += pkt_len;
        cl->rxLen -= pkt_len;
    }

    if (p != cl->rxBuffer && cl->rxLen > 0) {
        memmove(cl->rxBuffer, p, cl->rxLen);
    }
}

static void *poll_thread(void *arg)
{
    struct dS485_server *server = arg;

    while (server->active) {
        dS485dClient* cl;
        time_t now;

        now = time(NULL);
        pthread_rwlock_rdlock(&server->clientListRWLock);
        LL_FOREACH(server->clSessions, cl) {
            pthread_mutex_lock(&cl->lock);

            if (cl->fd < 0) {
                if (now - cl->ts_detach > CLIENT_INACTIVE_THRESHOLD) {
                    report2(LOG_WARNING, cl->peer_name, "disconnecting %d sec\n",
                            now - cl->ts_detach);
                }
            }

            if (now - cl->ts_activity > CLIENT_INACTIVE_THRESHOLD) {
                report2(LOG_WARNING, cl->peer_name,
                        "receiver thread blocked for %d sec cmd:%d\n",
                        (now - cl->ts_activity), cl->last_command);
            }

            pthread_mutex_unlock(&cl->lock);
        }
        pthread_rwlock_unlock(&server->clientListRWLock);
        sleep(MAX_THREAD_SLEEP);
    }
    return NULL;
}

int ds485_server_listen(struct dS485_server* server, int port, bool listenGlobal)
{
    struct pollfd fds[18]; /* 1 listen socket; 16 clients; +1 for kicking clients */
    int rval, nfds;

    rval = passivesock(port, "tcp", 0, listenGlobal);
    if (rval < 0) {
        report(LOG_ERR, "ERROR opening network socket\n");
        return rval;
    }
    server->sfd = rval;

    rval = pthread_create(&server->poll_tid, NULL, poll_thread, server);
    if (rval) {
        report(LOG_ERR, "poll_thread start: %s\n", strerror(rval));
        close(server->sfd);
        return -rval;
    }

    memset(&fds, 0, sizeof(fds));
    fds[0].fd = server->sfd;
    fds[0].events = POLLIN;
    nfds = 1;

    while (server->active) {
        struct sockaddr fsin;
        socklen_t cnamelen;
        int cfd;

        /* socket is non-blocking, wait for client */
        rval = poll(fds, nfds, 1000);
        if (rval < 0) {
            rval = errno;
            if (errno == EINTR) {
                continue;
            }
            report(LOG_ERR, "poll: %s\n", strerror(rval));
            return -rval;
        } else if (rval == 0) {
            // timeout
            continue;
        }

        if (!server->active) {
            break;
        }

        if (fds[0].revents & (POLLRDHUP | POLLERR | POLLHUP | POLLNVAL)) {
            /* TODO remove file descriptor, trigger action, continue */
            report(LOG_WARNING, "poll: file descriptor failed 0x%04x\n",
                   fds[0].revents);
            rval = -EBADF;
            break;
        }

        assert(fds[0].revents == POLLIN);
        cnamelen = sizeof(fsin);
        cfd = accept(server->sfd, (struct sockaddr *)&fsin, &cnamelen);
        if (cfd < 0) {
            report(LOG_WARNING, "new client: accept failed %s", strerror(errno));
            continue;
        }

        dS485dClient* cls = ds485_server_create_client(cfd);
        if (!cls) {
            continue;
        }

        rval = ds485_server_append_client(server, cls);
        if (rval < 0) {
            ds485_server_free_client(cls);
            continue;
        }

        /* start service thread */
        rval = pthread_create(&cls->tid, NULL, sysServiceThread, cls);
        if (rval) {
            report(LOG_ERR, "Could not create client thread. %s\n",
                   strerror(rval));
            ds485_server_remove_client(server, cls);
            ds485_server_free_client(cls);
            if (rval == EAGAIN) {
                rval = 0;
                continue; /* client will retry */
            }
            server->active = false;
            rval = -rval;
            break;
        }
        pthread_detach(cls->tid);
    }

    report(LOG_INFO, "server: shutdown\n");
    close(server->sfd);
    return rval;
}
NETLIB_EXPORTED(ds485_server_listen);

void ds485_server_abort(struct dS485_server *server)
{
    /* all threads periodically poll this variable */
    server->active = false;

    pthread_join(server->poll_tid, NULL);

    /* TODO pthread_join */
    sleep(MAX_THREAD_SLEEP);
}
NETLIB_EXPORTED(ds485_server_abort);
